from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Generator, Optional, Tuple, Union

import pandas as pd
from cerberus import Validator

from kabutobashi.errors import KabutobashiEntityError

from .stock_data_processed import StockDataParameterized, StockDataProcessed


@dataclass(frozen=True)
class StockDataSingleDay:
    """

    * code: 銘柄コード
    * open: 始値
    * close: 終値
    * high: 高値
    * low: 底値
    * unit: 単位
    * volume: 出来高
    * per: PER
    * psr: PSR
    * pbr: PBR
    * market: 市場
    * dt: その株価の値の日
    * name: 名前
    * industry_type: 業種

    Args:
        code: 銘柄コード
        market: 市場
        industry_type: 業種
        open: 円
        high: 円
        low: 円
        close: 円
    """

    code: str
    market: str
    name: str
    industry_type: str
    open: float
    high: float
    low: float
    close: float
    psr: float
    per: float
    pbr: float
    volume: int
    unit: int
    market_capitalization: str
    issued_shares: str
    dt: str
    _SCHEMA = {
        "code": {"type": "string"},
        "market": {"type": "string"},
        "industry_type": {"type": "string"},
        "name": {"type": "string"},
        "open": {"type": "float"},
        "high": {"type": "float"},
        "low": {"type": "float"},
        "close": {"type": "float"},
        "psr": {"type": "float"},
        "per": {"type": "float"},
        "pbr": {"type": "float"},
        "volume": {"type": "integer"},
        "unit": {"type": "integer"},
        "market_capitalization": {"type": "string"},
        "issued_shares": {"type": "string"},
        "dt": {"type": "string"},
    }

    def __post_init__(self):
        validator = Validator(self._SCHEMA)
        if not validator.validate(self.dumps()):
            raise KabutobashiEntityError(validator)

    @staticmethod
    def schema() -> list:
        return list(StockDataSingleDay._SCHEMA.keys())

    @staticmethod
    def from_page_of(data: dict) -> "StockDataSingleDay":
        label_split = data["stock_label"].split("  ")
        return StockDataSingleDay(
            code=label_split[0],
            market=label_split[1],
            name=data["name"],
            industry_type=data["industry_type"],
            open=float(StockDataSingleDay._convert(data["open"])),
            high=float(StockDataSingleDay._convert(data["high"])),
            low=float(StockDataSingleDay._convert(data["low"])),
            close=float(StockDataSingleDay._convert(data["close"])),
            unit=int(StockDataSingleDay._convert(data["unit"])),
            psr=float(StockDataSingleDay._convert(data["psr"])),
            per=float(StockDataSingleDay._convert(data["per"])),
            pbr=float(StockDataSingleDay._convert(data["pbr"])),
            volume=int(StockDataSingleDay._convert(data["volume"])),
            market_capitalization=data["market_capitalization"],
            issued_shares=data["issued_shares"],
            dt=data["date"],
        )

    @staticmethod
    def _convert(input_value: str) -> str:
        return input_value.replace("---", "0").replace("円", "").replace("株", "").replace("倍", "").replace(",", "")

    def dumps(self) -> dict:
        return asdict(self)


@dataclass(frozen=True)
class StockDataSingleCode:
    """
    単一銘柄の複数日の株データを保持するEntity

    以下のデータを保持する

    * code: 銘柄コード
    * open: 始値
    * close: 終値
    * high: 高値
    * low: 底値
    * unit: 単位
    * volume: 出来高
    * per: PER
    * psr: PSR
    * pbr: PBR
    * market: 市場
    * dt: その株価の値の日
    * name: 名前
    * industry_type: 業種

    Args:
        df: 複数日・単一銘柄を保持するDataFrame
        code: 銘柄コード

    """

    df: pd.DataFrame
    code: str
    stop_updating: bool
    contains_outlier: bool
    REQUIRED_COL = ["code", "open", "close", "high", "low", "unit", "volume", "per", "psr", "pbr", "market", "dt"]
    OPTIONAL_COL = ["name", "industry_type"]

    def __post_init__(self):
        self._null_check()
        self._code_constraint_check(df=self.df)
        if not self._validate():
            raise KabutobashiEntityError(f"required: {self.REQUIRED_COL}, input: {self.df.columns}")

    def _null_check(self):
        if self.df is None:
            raise KabutobashiEntityError("required")

    def _validate(self) -> bool:
        columns = list(self.df.columns)
        # 必須のカラム確認
        if not all([item in columns for item in self.REQUIRED_COL]):
            return False
        return True

    @staticmethod
    def _code_constraint_check(df: pd.DataFrame):
        df_columns = df.columns
        if "code" in df_columns:
            code = list(set(df.code.values))
            if len(code) > 1:
                raise KabutobashiEntityError("multiple code")
            elif len(code) == 0:
                raise KabutobashiEntityError("no code")

    @staticmethod
    def of(df: pd.DataFrame):
        df_columns = df.columns
        # 日付カラムの候補値を探す
        date_column = None
        if "date" in df_columns:
            date_column = "date"
        elif "dt" in df_columns:
            date_column = "dt"
        elif "crawl_datetime" in df_columns:
            date_column = "crawl_datetime"
        if date_column is None:
            raise KabutobashiEntityError("日付のカラム[dt, date, crawl_datetime]のいずれかが存在しません")
        if ("date" in df_columns) and ("dt" in df_columns) and ("crawl_datetime" in df_columns):
            raise KabutobashiEntityError("日付のカラム[dt, date]は片方しか存在できません")

        # 変換
        if date_column == "crawl_datetime":
            df["dt"] = df["crawl_datetime"].apply(lambda x: datetime.fromisoformat(x).strftime("%Y-%m-%d"))
            date_column = "dt"
        # indexにdateを指定
        idx = pd.to_datetime(df[date_column]).sort_index()

        # codeの確認
        StockDataSingleCode._code_constraint_check(df=df)
        if "code" in df_columns:
            code = list(set(df.code.values))[0]
        else:
            code = "-"

        # 数値に変換・「業種」という文字列削除
        df = df.assign(
            open=df["open"].apply(StockDataSingleCode._replace_comma),
            close=df["close"].apply(StockDataSingleCode._replace_comma),
            high=df["high"].apply(StockDataSingleCode._replace_comma),
            low=df["low"].apply(StockDataSingleCode._replace_comma),
            pbr=df["pbr"].apply(StockDataSingleCode._replace_comma),
            psr=df["psr"].apply(StockDataSingleCode._replace_comma),
            per=df["per"].apply(StockDataSingleCode._replace_comma),
        )
        if "industry_type" in df_columns:
            df["industry_type"] = df["industry_type"].apply(lambda x: x.replace("業種", ""))

        df.index = idx
        df = df.fillna(0)
        df = df.convert_dtypes()
        # order by dt
        df = df.sort_index()
        return StockDataSingleCode(
            code=code,
            df=df,
            stop_updating=StockDataSingleCode._check_recent_update(df=df),
            contains_outlier=StockDataSingleCode._check_outlier_value(df=df),
        )

    @staticmethod
    def _replace_comma(x) -> float:
        """
        pandas内の値がカンマ付きの場合に、カンマを削除する関数
        :param x:
        :return:
        """
        if type(x) is str:
            x = x.replace(",", "")
        try:
            f = float(x)
        except ValueError as e:
            raise KabutobashiEntityError(f"floatに変換できる値ではありません。{e}")
        return f

    @staticmethod
    def _check_recent_update(df: pd.DataFrame) -> bool:
        """
        直近の更新が止まっているかどうか
        """
        return (
            (len(df["open"].tail(10).unique()) == 1)
            or (len(df["high"].tail(10).unique()) == 1)
            or (len(df["low"].tail(10).unique()) == 1)
            or (len(df["close"].tail(10).unique()) == 1)
        )

    @staticmethod
    def _check_outlier_value(df: pd.DataFrame) -> bool:
        """
        不正な値が含まれている場合にtrueを返す

        以下の場合にTrueになる
        - 急に0になる値が含まれている
        """
        return (
            (len(df[df["open"] == 0].index) > 0)
            or (len(df[df["high"] == 0].index) > 0)
            or (len(df[df["low"] == 0].index) > 0)
            or (len(df[df["close"] == 0].index) > 0)
        )

    def sliding_split(
        self, *, buy_sell_term_days: int = 5, sliding_window: int = 60, step: int = 3
    ) -> Generator[Tuple[int, pd.DataFrame, pd.DataFrame], None, None]:
        """
        単一の銘柄に関してwindow幅を ``sliding_window`` 日として、
        保持しているデータの期間の間をslidingしていく関数。

        Args:
            buy_sell_term_days: この日数後までデータを切り出す。
            sliding_window: slidingしていくwindow幅
            step: windowsをずらしていく期間

        Returns:
            idx: 切り出された番号。
            df_for_x: 特徴量を計算するためのDataFrame。
            df_for_y: `buy_sell_term_days`後のDataFrameを返す。値動きを追うため。
        """
        df_length = len(self.df.index)
        if df_length < buy_sell_term_days + sliding_window:
            raise KabutobashiEntityError("入力されたDataFrameの長さがwindow幅よりも小さいです")
        loop = df_length - (buy_sell_term_days + sliding_window)
        for idx, i in enumerate(range(0, loop, step)):
            offset = i + sliding_window
            end = offset + buy_sell_term_days
            yield idx, self.df[i:offset], self.df[offset:end]

    def get_df(self, minimum=True, latest=False):
        df = self.df

        if latest:
            latest_dt = max(df["dt"])
            df = df[df["dt"] == latest_dt]

        if minimum:
            return df[self.REQUIRED_COL]
        else:
            return df[self.REQUIRED_COL + self.OPTIONAL_COL]

    def to_processed(self, methods: list) -> StockDataProcessed:
        return StockDataProcessed.of(df=self.df, methods=methods)

    def to_parameterize(self, methods: list, buy_sell_term_days: int = 5) -> StockDataParameterized:
        df_x = self.df[:-buy_sell_term_days]
        df_y = self.df[-buy_sell_term_days:]
        return StockDataParameterized.of(df_x=df_x, df_y=df_y, methods=methods)


@dataclass(frozen=True)
class StockDataMultipleCode:
    """
    複数銘柄の複数日の株データを保持するEntity

    単一銘柄のデータのみを返したり、複数銘柄のデータをループで取得できるクラス。

    Args:
        df: 複数日・複数銘柄を保持するDataFrame

    Examples:
        >>> import kabutobashi as kb
        >>> sdmc = kb.example()
        >>> sdsc = sdmc.to_single_code(code=1375)
    """

    df: pd.DataFrame
    REQUIRED_COL = StockDataSingleCode.REQUIRED_COL
    OPTIONAL_COL = StockDataSingleCode.OPTIONAL_COL

    def __post_init__(self):
        self._null_check()
        if not self._validate():
            raise KabutobashiEntityError(f"不正なデータ構造です: {self.df.columns=}")

    def _null_check(self):
        if self.df is None:
            raise KabutobashiEntityError("required")

    def _validate(self) -> bool:
        columns = list(self.df.columns)
        # 必須のカラム確認
        if not all([item in columns for item in self.REQUIRED_COL]):
            return False
        return True

    @staticmethod
    def of(df: pd.DataFrame) -> "StockDataMultipleCode":
        return StockDataMultipleCode(df=df)

    def to_single_code(self, code: Union[str, int]) -> StockDataSingleCode:
        return StockDataSingleCode.of(df=self.df[self.df["code"] == code])

    def to_code_iterable(
        self,
        until: Optional[int] = None,
        *,
        skip_reit: bool = True,
        row_more_than: Optional[int] = 80,
        code_list: list = None,
    ) -> Generator[StockDataSingleCode, None, None]:
        _count = 0
        df = self.df.copy()

        if code_list:
            df = df[df["code"].isin(code_list)]
        if skip_reit:
            df = df[~(df["market"] == "東証REIT")]

        for code, df_ in df.groupby("code"):
            if row_more_than:
                if len(df_.index) < row_more_than:
                    continue
            if until:
                if _count >= until:
                    return
            _count += 1

            sdsc = StockDataSingleCode.of(df=df_)
            if sdsc.stop_updating:
                continue
            if sdsc.contains_outlier:
                continue
            yield sdsc

    def get_df(self, minimum=True, latest=False, code_list: list = None):
        df = self.df

        if code_list:
            df = df[df["code"].isin(code_list)]
        if latest:
            latest_dt = max(df["dt"])
            df = df[df["dt"] == latest_dt]

        if minimum:
            return df[self.REQUIRED_COL]
        else:
            return df[self.REQUIRED_COL + self.OPTIONAL_COL]

    @staticmethod
    def read():
        from kabutobashi.repository.stock_data_repository import StockDataMultipleCodeReader

        return StockDataMultipleCodeReader()

    def write(self):
        from kabutobashi.repository.stock_data_repository import StockDataMultipleCodeWriter

        return StockDataMultipleCodeWriter(multiple_code=self)



import kabutobashi as kb


def test_crawl_page_detail():
    result = kb.StockInfoPage(code=4395).get()
    assert result is not None
    assert type(result) is dict


def test_crawl_ipo_list():
    result = kb.StockIpoPage(year=2019).get()
    assert result is not None
    assert type(result) is dict


def test_crawl_week_52_high_low_list():
    result = kb.Weeks52HighLowPage(data_type="high").get()
    assert result is not None
    assert type(result) is dict



from dataclasses import dataclass
from typing import Union

from bs4 import BeautifulSoup

from kabutobashi.domain.entity import StockIpo

from .page import Page


@dataclass(frozen=True)
class StockIpoPage(Page):
    year: Union[str, int]
    base_url: str = "https://96ut.com/ipo/list.php"

    def url(self) -> str:
        if type(self.year) is int:
            year = str(self.year)
        elif type(self.year) is str:
            year = self.year
        else:
            raise ValueError("")
        return f"{self.base_url}?year={year}"

    def _get(self) -> dict:
        res = BeautifulSoup(self.get_url_text(url=self.url()), features="lxml")
        table_content = res.find("div", {"class": "tablewrap"})
        table_thead = table_content.find("thead")
        # headの取得
        table_head_list = []
        for th in table_thead.find_all("th"):
            table_head_list.append(th.get_text())

        # bodyの取得
        table_tbody = table_content.find("tbody")
        whole_result = []
        for idx, tr in enumerate(table_tbody.find_all("tr")):
            table_body_dict = {}
            for header, td in zip(table_head_list, tr.find_all("td")):
                table_body_dict[header] = td.get_text().replace("\n", "")
            whole_result.append(StockIpo.from_page_of(data=table_body_dict).dumps())
        return {"ipo_list": whole_result}



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cerberus import Validator
from mplfinance.original_flavor import candlestick_ohlc

from kabutobashi.errors import KabutobashiEntityError


@dataclass(frozen=True)
class StockDataProcessed:
    """
    methodで処理した後のデータを保持
    可視化などを実行する際に利用

    Examples:
        >>> import kabutobashi as kb
        >>> sdmc = kb.example()
        >>> sdp = sdmc.to_single_code(1375).to_processed([kb.sma, kb.macd])
        >>> data = sdp.get_impact()
        {"sma": 0.4, "macd": -0.04}
        >>> df = pd.read_csv(...)
        >>> sma = kb.SMA(short_term=3, medium_term=15, long_term=50)
        >>> processed = kb.StockDataProcessed.of(df=df, methods=[sma, kb.macd])
        >>> processed.get_impact()
        {"sma": 0.2, "macd": -0.04}

    """

    code: Optional[Union[str, int]]
    base_df: pd.DataFrame = field(repr=False)
    # {"method": "", "data": pd.DataFrame, "color_mapping": List[dict], "visualize_option": dict}
    processed_dfs: List[Dict[str, Any]] = field(repr=False)

    REQUIRED_DF_COLUMNS = ["code", "open", "close", "high", "low", "dt"]
    PROCESSED_SCHEMA = {
        "method": {"type": "string"},
        "data": {"required": True},
        "color_mapping": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "df_key": {"type": "string"},
                    "color": {"type": "string"},
                    "label": {"type": "string"},
                    "plot": {"type": "string", "allowed": ["plot", "bar"], "required": False},
                },
            },
        },
        "visualize_option": {"type": "dict", "schema": {"position": {"type": "string", "allowed": ["in", "lower"]}}},
    }

    def __post_init__(self):
        # pd.DataFrameの確認
        # self._validate()
        pass

    @staticmethod
    def of(df: pd.DataFrame, methods: list) -> "StockDataProcessed":
        from kabutobashi.domain.method import Method

        # check all methods
        for method in methods:
            if not isinstance(method, Method):
                raise KabutobashiEntityError()

        initial_method: Method = methods[0]
        rest_methods: List[Method] = methods[1:]
        base = initial_method.process(df=df)

        for rest_method in rest_methods:
            base = base + rest_method.process(df=df)
        return base

    def _validate(self):
        columns = list(self.base_df.columns)
        # 必須のカラム確認
        if not all([item in columns for item in self.REQUIRED_DF_COLUMNS]):
            return KabutobashiEntityError(columns)

        validator = Validator(self.PROCESSED_SCHEMA)
        if not validator.validate(self.processed_dfs):
            raise KabutobashiEntityError(validator)

    def __add__(self, other: "StockDataProcessed") -> "StockDataProcessed":
        if not isinstance(other, StockDataProcessed):
            raise KabutobashiEntityError()

        # update
        processed_dfs = []
        processed_dfs.extend(self.processed_dfs)
        processed_dfs.extend(other.processed_dfs)
        return StockDataProcessed(
            code=self.code,
            base_df=self.base_df,
            processed_dfs=processed_dfs,
        )

    def get_impact(self, influence: int = 2, tail: int = 5) -> dict:
        """

        Args:
            influence:
            tail:

        Returns:
            Dict[str, float]

        Examples:
            >>> import kabutobashi as kb
            >>> df = kb.read_csv(...)
            >>> processed = kb.StockDataProcessed.of(_df=df, methods=[kb.sma, kb.macd])
            >>> processed.get_impact()
            {"sma": 0.4, "macd": -0.04}
            >>> sma = kb.SMA(short_term=3, medium_term=15, long_term=50)
            >>> processed = kb.StockDataProcessed.of(_df=df, methods=[sma, kb.macd])
            >>> processed.get_impact()
            {"sma": 0.2, "macd": -0.04}
        """
        result_dict = {}
        for v in self.processed_dfs:
            result_dict.update({v["method"]: self._get_impact(_df=v["data"], influence=influence, tail=tail)})
        return result_dict

    @staticmethod
    def _get_impact(_df: pd.DataFrame, influence: int, tail: int) -> float:
        """
        売りと買いのシグナルの余波の合計値を返す。

        Args:
            _df:
            influence:
            tail:

        Returns:
            [-1,1]の値をとる。-1: 売り、1: 買いを表す
        """
        _df["buy_impact"] = _df["buy_signal"].ewm(span=influence).mean()
        _df["sell_impact"] = _df["sell_signal"].ewm(span=influence).mean()
        buy_impact_index = _df["buy_impact"].iloc[-tail:].sum()
        sell_impact_index = _df["sell_impact"].iloc[-tail:].sum()
        return round(buy_impact_index - sell_impact_index, 5)

    @staticmethod
    def _add_ax_candlestick(ax, _df: pd.DataFrame):
        # datetime -> float
        time_series = mdates.date2num(_df["dt"])
        data = _df[["open", "high", "low", "close"]].values.T
        # data
        ohlc = np.vstack((time_series, data)).T
        candlestick_ohlc(ax, ohlc, width=0.7, colorup="g", colordown="r")

    def visualize(self, size_ratio: int = 2):
        """
        macdはlower
        sma、bolinger_bandsは同じところに表示させる。
        買いのポイントも表示させる

        Returns:

        """

        def _n_rows() -> int:
            lower_nums = len([p for p in self.processed_dfs if p["visualize_option"]["position"] == "lower"])
            return 1 + lower_nums

        n_rows = _n_rows()

        def _gridspec_kw() -> dict:
            if n_rows == 1:
                return {"height_ratios": [3]}
            return {"height_ratios": [3] + [1] * (n_rows - 1)}

        gridspec_kw = _gridspec_kw()
        fig, axs = plt.subplots(
            nrows=n_rows, ncols=1, figsize=(6 * size_ratio, 5 * size_ratio), gridspec_kw=gridspec_kw
        )
        # auto-formatting x-axis
        fig.autofmt_xdate()
        # set candlestick base
        self._add_ax_candlestick(axs[0], self.base_df)

        ax_idx = 1
        # plots
        for processed in self.processed_dfs:
            position = processed["visualize_option"]["position"]
            df = processed["data"]
            time_series = mdates.date2num(df["dt"])
            mapping = processed["color_mapping"]
            if position == "in":
                for m in mapping:
                    df_key = m["df_key"]
                    color = m["color"]
                    label = m["label"]
                    axs[0].plot(time_series, df[df_key], label=label)
                # display labels
                axs[0].legend(loc="best")
            elif position == "lower":
                for m in mapping:
                    df_key = m["df_key"]
                    color = m["color"]
                    label = m["label"]
                    plot = m.get("plot", "plot")
                    if plot == "plot":
                        # type FloatingArray is no accepted ...
                        # so `df[df_key].astype(float)`
                        axs[ax_idx].plot(time_series, df[df_key].astype(float), label=label)
                    elif plot == "bar":
                        axs[ax_idx].bar(time_series, df[df_key], label=label)
                # display labels
                axs[ax_idx].legend(loc="best")
                # lower
                ax_idx += 1
            elif position == "-":
                # technical_analysis以外のmethodが入っている場合
                pass
            else:
                raise KabutobashiEntityError()

        return fig


@dataclass(frozen=True)
class StockDataParameterized:
    """
    Examples:
        >>> import kabutobashi as kb
        >>> import pandas as pd
        >>> data_list = []
        >>> sdmc = kb.example()
        >>> parameterize_methods = kb.methods + [kb.basic]
        >>> for sdsc in sdmc.to_code_iterable(until=1, row_more_than=80):
        ...     code = sdsc.code
        ...     for idx, df_x, df_y in sdsc.sliding_split():
        ...         df_params = kb.StockDataParameterized.of(df_x=df_x, df_y=df_y, methods=parameterize_methods)
        ...         print(f"code:{code}, x:{df_params.x()}, y:{df_params.y()}")
        ...         data_list.append(df_params.row())
        >>> data_for_ml = pd.DataFrame(data_list)
    """

    start_at: str
    end_at: str
    days_after_n: int
    day_after_diff: float
    code: str
    parameters: Dict[str, float]

    @staticmethod
    def of(df_x: pd.DataFrame, df_y: pd.DataFrame, methods: list) -> "StockDataParameterized":
        from kabutobashi.domain.method import Method

        # check all methods
        for method in methods:
            if not isinstance(method, Method):
                raise KabutobashiEntityError()

        initial_method: Method = methods[0]
        rest_methods: List[Method] = methods[1:]
        base = initial_method.parameterize(df_x=df_x, df_y=df_y)

        for rest_method in rest_methods:
            base = base + rest_method.parameterize(df_x=df_x, df_y=df_y)
        return base

    def __add__(self, other: "StockDataParameterized") -> "StockDataParameterized":
        if not isinstance(other, StockDataParameterized):
            raise KabutobashiEntityError()

        # update
        params = {}
        params.update(self.parameters)
        params.update(other.parameters)
        return StockDataParameterized(
            code=self.code,
            start_at=self.start_at,
            end_at=self.end_at,
            days_after_n=self.days_after_n,
            day_after_diff=self.day_after_diff,
            parameters=params,
        )

    def x(self) -> dict:
        return self.parameters

    def y(self) -> float:
        return self.day_after_diff

    def row(self) -> dict:
        row_ = {}
        row_.update(self.x())
        row_.update({"diff": self.y()})
        return row_



import pandas as pd
import pytest

import kabutobashi as kb


class TestStockInfo:
    def test_error_init(self):
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.StockDataSingleDay(
                code="1234",
                market="market",
                name="",
                industry_type="industry_type",
                open="",
                high="",
                low="",
                close="",
                psr="",
                per="",
                pbr="",
                volume="",
                unit="",
                market_capitalization="",
                issued_shares="",
                dt="",
            )


class TestStockIpo:
    def test_error_init(self):
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.StockIpo(
                code="", market="", manager="", stock_listing_at="", public_offering="", evaluation="", initial_price=""
            )


class TestWeeks52HihLow:
    def test_error_init(self):
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.Weeks52HighLow(
                code="", brand_name="", close="", buy_or_sell="", volatility_ratio="", volatility_value=""
            )


class TestStockDataSingleCode:
    def test_of(self, data_path):
        df = pd.read_csv(f"{data_path}/example.csv.gz")
        single_code = df[df["code"] == 1375]
        _ = kb.StockDataSingleCode.of(df=single_code)

        # check None
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.StockDataSingleCode(code="-", df=None, stop_updating=False, contains_outlier=False)

        # check multiple code
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.StockDataSingleCode(code="-", df=df, stop_updating=False, contains_outlier=False)

        # check invalid column
        with pytest.raises(kb.errors.KabutobashiEntityError):
            _ = kb.StockDataSingleCode(code="-", df=single_code[["close"]], stop_updating=False, contains_outlier=False)

    def test_get_df(self, data_path):
        df = pd.read_csv(f"{data_path}/example.csv.gz")
        single_code = df[df["code"] == 1375]
        sdsc = kb.StockDataSingleCode.of(df=single_code)

        required_cols = kb.StockDataSingleCode.REQUIRED_COL
        optional_cols = kb.StockDataSingleCode.OPTIONAL_COL

        # check minimum df
        minimum_df = sdsc.get_df()
        assert all([(c in minimum_df.columns) for c in required_cols])
        assert all([(c not in minimum_df.columns) for c in optional_cols])

        # check full df
        full_df = sdsc.get_df(minimum=False)
        assert all([(c in full_df.columns) for c in required_cols])
        assert all([(c in full_df.columns) for c in optional_cols])

        latest_date_df = sdsc.get_df(latest=True)
        assert len(latest_date_df.index) == 1


class TestStockDataProcessed:
    def test_of(self):
        sdmc = kb.example()
        sdsc = sdmc.to_single_code(code=1375)
        processed = kb.StockDataProcessed.of(df=sdsc.df, methods=[kb.sma, kb.macd])
        _ = processed.visualize()


class TestStockDataParameterized:
    def test_of(self):
        sdmc = kb.example()
        sdsc = sdmc.to_single_code(code=1375)

        methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility, kb.industry_cat]
        for idx, df_x, df_y in sdsc.sliding_split():
            parameterized = kb.StockDataParameterized.of(df_x=df_x, df_y=df_y, methods=methods)
            assert type(parameterized.x()) is dict
            assert type(float(parameterized.y())) is float
            assert type(parameterized.row()) is dict



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Generator, Union

import pandas as pd

from kabutobashi.domain.entity import StockDataMultipleCode
from kabutobashi.utilities import get_past_n_days

__all__ = ["StockDataMultipleCodeReader", "StockDataMultipleCodeWriter"]


class IStockDataMultipleCodeReader(metaclass=ABCMeta):
    @abstractmethod
    def _path(self) -> Generator[str, None, None]:
        raise NotImplementedError()

    def read(self) -> StockDataMultipleCode:
        return self._read()

    def _read(self) -> StockDataMultipleCode:
        df = pd.concat([pd.read_csv(p) for p in self._path()])
        return StockDataMultipleCode.of(df=df)


@dataclass(frozen=True)
class StockDataMultipleCodeBasicReader(IStockDataMultipleCodeReader):
    path_candidate: Union[str, list]

    def _path(self) -> Generator[str, None, None]:
        if type(self.path_candidate) is str:
            yield self.path_candidate
        elif type(self.path_candidate) is list:
            for path in self.path_candidate:
                yield path
        else:
            raise ValueError()


@dataclass(frozen=True)
class StockDataMultipleCodeTargetDateReader(IStockDataMultipleCodeReader):
    path_format: str
    start_date: str
    n: int

    def _path(self) -> Generator[str, None, None]:
        date_list = get_past_n_days(current_date=self.start_date, n=self.n)
        path_list = [self.path_format.format(dt=dt) for dt in date_list]

        for path in path_list:
            yield path


class IStockDataMultipleCodeWriter(metaclass=ABCMeta):
    @abstractmethod
    def _path(self) -> Generator[str, None, None]:
        raise NotImplementedError()

    def write(self, stock_data_multiple_code: StockDataMultipleCode):
        return self._write(stock_data_multiple_code=stock_data_multiple_code)

    @abstractmethod
    def _write(self, stock_data_multiple_code: StockDataMultipleCode):
        raise NotImplementedError()


@dataclass(frozen=True)
class StockDataMultipleCodeBasicWriter(IStockDataMultipleCodeWriter):
    path_candidate: Union[str, list]

    def _path(self) -> Generator[str, None, None]:
        if type(self.path_candidate) is str:
            yield self.path_candidate
        elif type(self.path_candidate) is list:
            for path in self.path_candidate:
                yield path
        else:
            raise ValueError()

    def _write(self, stock_data_multiple_code: StockDataMultipleCode):
        # zip()をyieldでも利用できる？
        for p in self._path():
            stock_data_multiple_code.df.to_csv(p, index=False)


class StockDataMultipleCodeReader:
    @staticmethod
    def csv(path_candidate: Union[str, list]) -> StockDataMultipleCode:
        return StockDataMultipleCodeBasicReader(path_candidate=path_candidate).read()

    @staticmethod
    def csv_from_past_n_days(path_format: str, start_date: str, n: int) -> StockDataMultipleCode:
        return StockDataMultipleCodeTargetDateReader(path_format=path_format, start_date=start_date, n=n).read()


@dataclass
class StockDataMultipleCodeWriter:
    multiple_code: StockDataMultipleCode

    def csv(self, path_candidate: str):
        return StockDataMultipleCodeBasicWriter(path_candidate=path_candidate).write(
            stock_data_multiple_code=self.multiple_code
        )



"""
kabutobashiにおいて、株に関する操作をする際に基本となるデータ構造を保持する。

- データ収集時に主に利用

  - StockIpo
  - Weeks52HighLow

- 分析時に主に利用

  - StockDataSingleDay
  - StockDataSingleCode
  - StockDataMultipleCode
  - StockDataProcessed
  - StockDataParameterized
"""
from .stock_data_processed import StockDataParameterized, StockDataProcessed
from .stock_data_raw import StockDataMultipleCode, StockDataSingleCode, StockDataSingleDay
from .stock_ipo import StockIpo
from .weeks_52_high_low_info import Weeks52HighLow



from dataclasses import dataclass
from typing import Union

from bs4 import BeautifulSoup

from kabutobashi.domain.entity import StockDataSingleDay

from .page import Page, PageDecoder


@dataclass(frozen=True)
class StockInfoPage(Page):
    """

    Examples:
        >>> sip = StockInfoPage(code="0001")
        >>> result = sip.get()
    """

    code: Union[int, str]
    base_url: str = "https://minkabu.jp/stock/{code}"

    def url(self) -> str:
        return self.base_url.format(code=self.code)

    def _get(self) -> dict:
        soup = BeautifulSoup(self.get_url_text(url=self.url()), features="lxml")
        result = {}

        stock_board_tag = "ly_col ly_colsize_7 md_box ly_row ly_gutters"

        # ページ上部の情報を取得
        stock_board = soup.find("div", {"class": stock_board_tag})
        result.update(
            {
                "stock_label": PageDecoder(tag1="div", class1="stock_label").decode(bs=stock_board),
                "name": PageDecoder(tag1="p", class1="md_stockBoard_stockName").decode(bs=stock_board),
                "close": PageDecoder(tag1="div", class1="stock_price").decode(bs=stock_board),
                "date": PageDecoder(tag1="h2", class1="stock_label fsl").decode(bs=stock_board),
            }
        )

        # ページ中央の情報を取得
        stock_detail = soup.find("div", {"class": "stock-detail"})
        info = {}
        for li in stock_detail.find_all("li", {"class": "ly_vamd"}):
            info[li.find("dt").get_text()] = li.find("dd").get_text()
        result.update(
            {
                "industry_type": PageDecoder(tag1="div", class1="ly_content_wrapper size_ss").decode(bs=stock_detail),
                "open": info["始値"],
                "high": info["高値"],
                "low": info["安値"],
                "unit": info["単元株数"],
                "per": info["PER(調整後)"],
                "psr": info["PSR"],
                "pbr": info["PBR"],
                "volume": info["出来高"],
                "market_capitalization": info["時価総額"],
                "issued_shares": info["発行済株数"],
            }
        )
        return StockDataSingleDay.from_page_of(data=result).dumps()
